#include "stdafx.h"
#include ".\outputmatroska.h"


#define CLUSTER_SPLIT 5000000000

// add to the end of the linked list
void OutputMatroska::addToBuffer(void* data, int dataSize, bool keyframe)
{
	FrameList* add = (FrameList*)malloc(sizeof(FrameList));
    
    add->data = new binary[dataSize];
    memcpy(add->data,data,dataSize);
    add->next=NULL;
	add->keyframe = keyframe;
    add->dataSize = dataSize;
	add->timecode = CurrentTimeCode;
	add->dataBuffer = new DataBuffer((binary *)add->data, dataSize);
    if(frameListStart == NULL)
    {
        // First element
        frameListStart = add;
    } else {
        frameListEnd->next = add;
    }
    frameListEnd = add;

}

// ----------------------------------------------------------------------------
void OutputMatroska::flushCluster()
{
	CurrentCluster = new KaxCluster();
	CurrentCluster->SetParent(FileSegment); // mandatory to store references in this Cluster
	CurrentCluster->SetPreviousTimecode(frameListStart->timecode,TIMECODE_SCALE); // the first timecode here
	lastClusterTime = frameListStart->timecode;
	FrameList* frameListCurrent=frameListStart;
   while(frameListCurrent)
    {
		KaxBlockGroup *MyNewBlock = NULL, *MyLastBlockTrk1 = NULL;

		if (frameListCurrent->keyframe) {
			CurrentCluster->AddFrame(*MyTrack1, frameListCurrent->timecode, *frameListCurrent->dataBuffer, MyNewBlock);
		} else {
			CurrentCluster->AddFrame(*MyTrack1, frameListCurrent->timecode, *frameListCurrent->dataBuffer, MyNewBlock, *MyLastBlockTrk1);
		}
		Cues.AddBlockGroup(*MyNewBlock);
	

		if (MyNewBlock != NULL) {
			MyLastBlockTrk1 = MyNewBlock;
		} else {
			MyLastBlockTrk1->SetBlockDuration(frameTime);
		}
        frameListCurrent = frameListCurrent->next;
    }
	

	ClusterSize += CurrentCluster->Render(*out_file,Cues,bWriteDefaultValues);
	CurrentCluster->ReleaseFrames();
	MetaSeek.IndexThis(*CurrentCluster, FileSegment);

	delete CurrentCluster;
	CurrentCluster = NULL;
	frameListCurrent=frameListStart;
    while(frameListCurrent)
    {
        FrameList* frameListOld = frameListCurrent;
        frameListCurrent = frameListCurrent->next;
		//delete frameListOld->dataBuffer;
		delete frameListOld->data;
		free(frameListOld);
    }
    frameListStart=(FrameList*)NULL;
    frameListEnd=(FrameList*)NULL;
}
// ----------------------------------------------------------------------------



OutputMatroska::~OutputMatroska(void)
{
	flushCluster();
	uint32 CueSize = Cues.Render(*out_file, bWriteDefaultValues);
	MetaSeek.IndexThis(Cues, FileSegment);
	
	/*MetaSeekSize =*/ Dummy.ReplaceWith(MetaSeek, *out_file, bWriteDefaultValues);
	if (FileSegment.ForceSize(SegmentSize - FileSegment.HeadSize() + MetaSeekSize + TrackSize + ClusterSize + CueSize + InfoSize)) {
		FileSegment.OverwriteHead(*out_file);
	}

	out_file->close();
	delete out_file;
}

void OutputMatroska::writeData(void* data, int dataSize, bool keyframe)
{
	if (CurrentTimeCode+frameTime > lastClusterTime+CLUSTER_SPLIT) {
		//flushCluster();
	}

	addToBuffer(data,dataSize,keyframe);

	CurrentTimeCode+=frameTime;
}

OutputMatroska::OutputMatroska(char* Filename, int width, int height, DWORD length, DWORD rate, DWORD scale, int bmilength,BITMAPINFOHEADER* bmi)
{
	frameListStart=NULL;
	frameListEnd=NULL;
	CurrentCluster=NULL;
	CurrentTimeCode = 0;
	ClusterSize = 0;
	lastClusterTime=0;
	frameTime = (scale*1000*TIMECODE_SCALE)/(rate);
	//frameTime = (scale*1000)/(rate*TIMECODE_SCALE);
    //try {
		// write the head of the file (with everything already configured)
		out_file = new StdIOCallback(Filename, MODE_CREATE);

		///// Writing EBML test
		EbmlHead FileHead;

		EDocType & MyDocType = GetChild<EDocType>(FileHead);
		*static_cast<EbmlString *>(&MyDocType) = "matroska";

		EDocTypeVersion & MyDocTypeVer = GetChild<EDocTypeVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeVer)) = 1;

		EDocTypeReadVersion & MyDocTypeReadVer = GetChild<EDocTypeReadVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeReadVer)) = 1;

		FileHead.Render(*out_file, bWriteDefaultValues);

		// size is unknown and will always be, we can render it right away
		SegmentSize = FileSegment.WriteHead(*out_file, 5, bWriteDefaultValues);

		KaxTracks & MyTracks = GetChild<KaxTracks>(FileSegment);

		// reserve some space for the Meta Seek writen at the end		
		Dummy.SetSize(30000); // 300 octets ask mosu about how much to reserve
		MetaSeekSize = Dummy.Render(*out_file, false);



		// fill the mandatory Info section
		KaxInfo & MyInfos = GetChild<KaxInfo>(FileSegment);
		KaxTimecodeScale & TimeScale = GetChild<KaxTimecodeScale>(MyInfos);
		*(static_cast<EbmlUInteger *>(&TimeScale)) = TIMECODE_SCALE;

		KaxDuration & SegDuration = GetChild<KaxDuration>(MyInfos);
		*(static_cast<EbmlFloat *>(&SegDuration)) = (double)(length*frameTime)/1000000;

		*((EbmlUnicodeString *)&GetChild<KaxMuxingApp>(MyInfos)) = L"libmatroska 0.4.4";
		*((EbmlUnicodeString *)&GetChild<KaxWritingApp>(MyInfos)) = L"avs2avi";
		GetChild<KaxWritingApp>(MyInfos).SetDefaultSize(25);
		InfoSize = MyInfos.Render(*out_file);
		MetaSeek.IndexThis(MyInfos, FileSegment);

		// fill track 1 params
		MyTrack1 = &GetChild<KaxTrackEntry>(MyTracks);
		MyTrack1->SetGlobalTimecodeScale(TIMECODE_SCALE);

		KaxTrackNumber & MyTrack1Number = GetChild<KaxTrackNumber>(*MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1Number)) = 1;

		KaxTrackUID & MyTrack1UID = GetChild<KaxTrackUID>(*MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1UID)) = 7;

		/*KaxTrackDefaultDuration & MyTrack1DefDur = GetChild<KaxTrackDefaultDuration>(*MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1DefDur)) = frameTime*TIMECODE_SCALE;*/


		*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(*MyTrack1))) = track_video;

		KaxCodecID & MyTrack1CodecID = GetChild<KaxCodecID>(*MyTrack1);
		*static_cast<EbmlString *>(&MyTrack1CodecID) = "V_MS/VFW/FOURCC";

		KaxCodecPrivate & MyTrackCodecPrivate = GetChild<KaxCodecPrivate>(*MyTrack1);
		MyTrackCodecPrivate.CopyBuffer((binary*)bmi,bmilength);
		MyTrack1->EnableLacing(false);

		// video specific params
		KaxTrackVideo & MyTrack1Video = GetChild<KaxTrackVideo>(*MyTrack1);

		KaxVideoPixelHeight & MyTrack1PHeight = GetChild<KaxVideoPixelHeight>(MyTrack1Video);
		*(static_cast<EbmlUInteger *>(&MyTrack1PHeight)) = height;

		KaxVideoPixelWidth & MyTrack1PWidth = GetChild<KaxVideoPixelWidth>(MyTrack1Video);
		*(static_cast<EbmlUInteger *>(&MyTrack1PWidth)) = width;

		TrackSize = MyTracks.Render(*out_file, bWriteDefaultValues);

		MetaSeek.IndexThis(MyTracks, FileSegment);
		//Cues.SetGlobalTimeCodeScale(TIMECODE_SCALE);
}
