// ============================================================================
// AVS2AVI v1.31 (c) 2002 by Christophe PARIS
// ----------------------------------------------------------------------------
// e-mail : christophe.paris@free.fr
// Web : http://christophe.paris.free.fr
// ----------------------------------------------------------------------------
// Code largely inspired by DVD2AVI (Chia-chen Kuo) & VirtualDub (Avery Lee)
// Thanks to int 21h from the Doom9 forum for his contributions
// ============================================================================
#include <stdio.h>
#include <conio.h>
#include "stdafx.h"
#include "OutputAvi.h"
#include "OutputLog.h"
#include "OutputMatroska.h"
#include "OutputNull.h"


//major_v.minor_v for display purposes
#define major_v "1"
#define minor_v "31"

#define VIDEO_BUFFERS 2

#ifndef ABOVE_NORMAL_PRIORITY_CLASS
#define ABOVE_NORMAL_PRIORITY_CLASS 0x00008000; // Yes you should update your platform SDK.
#endif

char outputArg = 'a';
OutputFormat* out;
HINSTANCE g_hInst; // vdubstuff

PAVIFILE pSRCfile = NULL;
PAVISTREAM pSRCVideoStream = NULL;
AVISTREAMINFO SRCStreamInfo;
BITMAPINFOHEADER bmihVideoSource;

COMPVARS CompVars;
COMPVARS SecondCompVars;
void *SRCData[VIDEO_BUFFERS];   //Made this global because it makes threading
                                //way easier

const char* StateFilename = NULL;
const char* FourCC = NULL;
BOOL SaveState = FALSE;
BOOL LoadState = FALSE;
BOOL Quiet = FALSE;
BOOL UseFourCC = FALSE;
BOOL Stop = FALSE;
BOOL StopRead = FALSE;
unsigned int StartFrame = 0;

HANDLE hSemWaitRead, hSemWaitWrite;

int NbPass = 1;
int current_pass = 1;
int current_frame = 0;
unsigned int FrameRead = 0, FrameWritten = 0, KFCount = 0, DSTAVISize = 0;

// ----------------------------------------------------------------------------

void InitializeSource(char* filename);
void ChooseOneCompatibleCompressor(COMPVARS *cvar);
void ChooseCompressorOnFourCC(const char* FourCC);
void ChooseCompatibleCompressor();
void InitializeDestination(char* filename);
void Compress();
void ReadFrame();
void EncodeFrame();
void Cleanup();
void CleanupAndExit(const int returnCode);
void ShowLastWindowsError(const char *pFormat,...);

BOOL WriteCodecParam(FILE* fileState, COMPVARS* cvar);
BOOL ReadCodecParam(FILE* fileState, COMPVARS* cvar);
void SaveCodecParams(const char* filename);
void LoadCodecParams(const char* filename);

const char* FourCC2String(const DWORD fourcc);
char* TimeMs2String(const int timeMs, char* buff);



// ----------------------------------------------------------------------------

void Usage()
{
    printf("\tUsage: avs2avi avs_filename avi_destination_filename [flags]\n\n");
    printf("\t\t-s <file> : Save codec parameters to <file>\n");
    printf("\t\t-l <file> : Load codec parameters from <file>\n");
    printf("\t\t-P <Passes>\t  : Enable Multi-pass encoding mode,\n");
    printf("\t\t-q\t  : Enable 'quiet' mode for more speed\n");
    printf("\t\t-c <4cc>  : Use codec having <4cc> with default settings\n");
    printf("\t\t-p [0-3]  : Priority (0:idle, 1:normal, 2: above, 3:high)\n");
	printf("\t\t-o [format]  : Output Format (a : Avi (default), m: Matroska (mkv), l : Log)\n");

}

// ----------------------------------------------------------------------------



void EraseCurrentLine()
{
    printf("\r                                                                            \r");
}

// ----------------------------------------------------------------------------

BOOL ParseCommandLine(int argc, char* argv[])
{
    DWORD dwPriorityClass;

    if(argc > 2)
    {
        for(int i = 3; i < argc; i++)
        {
            int n = strlen(argv[i]);
            for(int j=1; j<n; j++)
            {
                switch(argv[i][j])
                {
                    case 'q':
                        Quiet = TRUE;
                        break;

                    case '2':
                        NbPass = 2;
                        break;

                    case 'l':
                        StateFilename = argv[i+1];
                        LoadState = TRUE;
                        i = i + 1;
                        break;

                    case 'c':
                        FourCC = argv[i+1];
                        UseFourCC = TRUE;
                        i = i + 1;
                        break;

                    case 's':
                        StateFilename = argv[i+1];
                        SaveState = TRUE;
                        i = i + 1;
                        break;

                    case 'p':
                        switch(argv[i+1][0])
                        {
                            case '0':
                                dwPriorityClass = IDLE_PRIORITY_CLASS;
                                break;
                            case '1':
                                dwPriorityClass = NORMAL_PRIORITY_CLASS;
                                break;
                            case '2':
                                dwPriorityClass = ABOVE_NORMAL_PRIORITY_CLASS;
                                break;
                            case '3':
                                dwPriorityClass = HIGH_PRIORITY_CLASS;
                                break;
                            default:
                                return FALSE;
                        }
                        SetPriorityClass(GetCurrentProcess(),dwPriorityClass);
                        i = i + 1;
                        break;
					case 'o':
						outputArg = argv[i+1][0];
						break;
					case 'P':
						NbPass = atoi(argv[i+1]);
						break;
                    default:
                        return FALSE;
                }
            }
        } 

        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

// ----------------------------------------------------------------------------

void ShowLastWindowsError(const char *pFormat,...)
{
    char ContextMsg[512];

    // Format the variable length parameter list
    va_list va;
    va_start(va, pFormat);
    vsprintf(ContextMsg, pFormat, va);
    va_end(va);

    // Get windows error msg
    LPVOID lpErrorMsg;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),          // Default language
        (LPTSTR) &lpErrorMsg,
        0,
        NULL
        );

    // Show complete msg
    printf("\n%s : %s\n", ContextMsg, lpErrorMsg);

    // Free the buffer.
    LocalFree( lpErrorMsg );
}

// ----------------------------------------------------------------------------

const char* FourCC2String(const DWORD fourcc)
{
    static char buff[5] = {0,0,0,0,0};
    char* fcc = (char*)&fourcc;
    for(int i=0; i < 4; i++)
        buff[i] = fcc[i];

    return buff;
}

// ----------------------------------------------------------------------------

char* TimeMs2String(const int timeMs, char* buff)
{
    int hh,mm,ss, ms;
    ms = timeMs % 1000;
    ss = timeMs / 1000;
    mm = (ss / 60) % 60;
    hh = ss / 3600;
    ss = ss % 60;

    sprintf(buff,"%02d:%02d:%02d.%03d", hh, mm, ss, ms);

    return buff;
}

// ----------------------------------------------------------------------------

void OnCtrlC(int signal)
{
    Stop = TRUE;
}

// ----------------------------------------------------------------------------

int main(int argc, char* argv[])
{
    printf("\nAVS2AVI v%s.%s (c) 2002,2003 Christophe PARIS (christophe.paris@free.fr)\n",major_v,minor_v);
    printf("Contributions by: int21h\n");
    printf("Additional code (c) 2002,2003 David Leatherdale (dave@leatherdale.net)\n\n");

    if(!ParseCommandLine(argc,argv))
    {
        Usage();
        return 1;
    }

    // Install Ctrl-C handler
    signal(SIGINT,OnCtrlC);


    for(current_pass = 1; current_pass <=  NbPass && ! Stop; current_pass++)
    {
        AVIFileInit();

        InitializeSource(argv[1]);

        if(UseFourCC)
        {
            ChooseCompressorOnFourCC(FourCC);
        }
        else
        {

            if(current_pass == 1)
            {
                if(LoadState)
                {
                    LoadCodecParams(StateFilename);
                }
                else
                {
                    ChooseCompatibleCompressor();
                    if(SaveState)
                    {
                        SaveCodecParams(StateFilename);
                    }
                }
            }

        }

        InitializeDestination(argv[2]);

        Compress();
        Cleanup();
        CompVars = SecondCompVars;
    }

    exit(0);
}

// ----------------------------------------------------------------------------

void Cleanup()
{
	
	if (current_pass != NbPass && Stop) { // if we didnt get to the last pass the second compressor needs to be closed
		ICCompressorFree(&SecondCompVars);
	} 

    ICCompressorFree(&CompVars);
	delete out;
    if(pSRCVideoStream)
        AVIStreamRelease(pSRCVideoStream);

    if(pSRCfile)
        AVIFileRelease(pSRCfile);

    AVIFileExit();
}

// ----------------------------------------------------------------------------

void CleanupAndExit(const int returnCode)
{
    Cleanup();
	getchar();
    exit(returnCode);
}

// ----------------------------------------------------------------------------

void ShowCompressorInfo(COMPVARS *cvar)
{
    char codecName[128];
    ICINFO icinfo;

    // Show compressor name and FOURCC
    ZeroMemory(&icinfo,sizeof(ICINFO));
    icinfo.dwSize = sizeof(ICINFO);
    if(ICGetInfo(cvar->hic, &icinfo, sizeof(ICINFO)))
    {
        WideCharToMultiByte(CP_ACP, 0, icinfo.szDescription, -1, codecName, 128, 0, 0);
        printf("  * Name: %s\n", codecName);
        printf("  * FOURCC: %s\n", FourCC2String(icinfo.fccHandler));
    }
}

// ----------------------------------------------------------------------------

void ChooseOneCompatibleCompressor(COMPVARS *cvar)
{
    // Choose a compatible compressor
    ZeroMemory(cvar,sizeof(COMPVARS));
    cvar->cbSize = sizeof(COMPVARS);
    if(ICCompressorChoose(NULL, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME,
        &bmihVideoSource, NULL, cvar, NULL) == FALSE)
    {
        fprintf(stderr,"\nOperation aborted by user.\n");
        CleanupAndExit(1);
    }
}

// ----------------------------------------------------------------------------

void ChooseCompatibleCompressor()
{
    // Choose a compatible compressor
    ChooseOneCompatibleCompressor(&CompVars);

    printf("Compressor:\n");
    ShowCompressorInfo(&CompVars);

    if(NbPass != 1)
    {
        // Choose a compatible compressor for second pass
        ChooseOneCompatibleCompressor(&SecondCompVars);

        // Check both compressor are the same
        if(CompVars.fccHandler != SecondCompVars.fccHandler)
        {
            printf("Second Pass Compressor:\n");
            ShowCompressorInfo(&SecondCompVars);
            fprintf(stderr,"\nFirst and Second Compressors don't match!\nTerminating...");
            CleanupAndExit(1);
        }
    }
}

// ----------------------------------------------------------------------------

void ChooseCompressorOnFourCC(const char* FourCC)
{
    ZeroMemory(&CompVars,sizeof(COMPVARS));
    CompVars.cbSize = sizeof(COMPVARS);
    CompVars.fccType = ICTYPE_VIDEO;
    CompVars.fccHandler = *(DWORD*)FourCC;
    CompVars.dwFlags = ICMF_COMPVARS_VALID;
    CompVars.hic = ICOpen(CompVars.fccType, CompVars.fccHandler, ICMODE_COMPRESS);
    if(CompVars.hic == NULL)
    {
        fprintf(stderr,"\nThere is no codec with FourCC = '%s' !\nTerminating...", FourCC);
        CleanupAndExit(1);
    }
    CompVars.lKey = 15;
    CompVars.lQ = ICQUALITY_DEFAULT;

    printf("Compressor:\n");
    ShowCompressorInfo(&CompVars);
}

// ----------------------------------------------------------------------------

void InitializeSource(char* filename)
{
    // Omit this info if we're doing second pass
    // We already saw it!
    if(current_pass == 1)
    {
        printf("Source:\n");
        printf("  * Filename: %s\n", filename);
    }
    // Open AVS file
    if (AVIFileOpen(&pSRCfile, filename, OF_SHARE_DENY_WRITE, 0L) != AVIERR_OK)
    {
        ShowLastWindowsError("AVIFileOpen failed, unable to open %s :", filename);
        CleanupAndExit(1);
    }

    // Get AVS video stream
    if (AVIFileGetStream(pSRCfile, &pSRCVideoStream, streamtypeVIDEO,0) != AVIERR_OK)
    {
        ShowLastWindowsError("AVIFileGetStream failed, unable to get video stream 0");
        CleanupAndExit(1);
    }

    LONG cbBMI = sizeof(BITMAPINFOHEADER);
    ZeroMemory(&bmihVideoSource, cbBMI);
    bmihVideoSource.biSize = cbBMI;
    if(AVIStreamReadFormat(pSRCVideoStream, 0, &bmihVideoSource, &cbBMI) != AVIERR_OK)
    {
        ShowLastWindowsError("AVIStreamReadFormat failed");
        CleanupAndExit(1);
    }

    if(current_pass == 1)
    {
        printf("  * Bit depth: %d\n", bmihVideoSource.biBitCount);
        switch(bmihVideoSource.biCompression)
        {
            case mmioFOURCC( 'Y','U','Y','2'):
            case mmioFOURCC( 'Y','V','1','2'):
                printf("  * FOURCC: %s\n", FourCC2String(bmihVideoSource.biCompression));
                break;
            case BI_RGB:
                printf("  * FOURCC: none (RGB)\n");
                break;
            default:
                printf("  * FOURCC: %s (not tested)\n", FourCC2String(bmihVideoSource.biCompression));
        }
    }

    ZeroMemory(&SRCStreamInfo, sizeof(AVISTREAMINFO));
    if(AVIStreamInfo(pSRCVideoStream, &SRCStreamInfo, sizeof(AVISTREAMINFO)) != AVIERR_OK)
    {
        ShowLastWindowsError("AVIStreamInfo failed");
        CleanupAndExit(1);
    }

    if(current_pass == 1)
    {
        printf("  * Frames: %d\n", SRCStreamInfo.dwLength);
        printf("  * Resolution: %dx%d\n", SRCStreamInfo.rcFrame.right,
            SRCStreamInfo.rcFrame.bottom);
    }
}

// ----------------------------------------------------------------------------

void InitializeDestination(char* filename)
{
    if(NbPass != current_pass)
    {
        printf("Destination:\n");
        printf("  * Filename: %s\n", filename);
		out = new OutputNull();
	} 
	else
	{
		switch (outputArg)
		{
			case 'a': {
				LONG formatSize;
				formatSize = ICCompressGetFormatSize(CompVars.hic, (void*)&bmihVideoSource);
				void* bmi = malloc(formatSize);
				memset(bmi, 0, formatSize);

				ICCompressGetFormat(CompVars.hic,(BITMAPINFOHEADER *)&bmihVideoSource,bmi);
				out = new OutputAvi(filename,bmihVideoSource.biWidth,bmihVideoSource.biHeight,SRCStreamInfo.dwLength, SRCStreamInfo.dwRate,SRCStreamInfo.dwScale,&CompVars,formatSize,(BITMAPINFOHEADER*)bmi);
				break;
			}
			case 'm': {
				LONG formatSize;
				formatSize = ICCompressGetFormatSize(CompVars.hic, (void*)&bmihVideoSource);
				void* bmi = malloc(formatSize);
				memset(bmi, 0, formatSize);

				ICCompressGetFormat(CompVars.hic,(BITMAPINFOHEADER *)&bmihVideoSource,bmi);

				out = new OutputMatroska(filename,bmihVideoSource.biWidth,bmihVideoSource.biHeight,SRCStreamInfo.dwLength, SRCStreamInfo.dwRate,SRCStreamInfo.dwScale,formatSize,(BITMAPINFOHEADER*)bmi);
				break;
			}
			case 'l': out = new OutputLog(filename);
				break;
			default:
				printf("Unrecognised Output Format Requested %c", outputArg);
				CleanupAndExit(1);
		}
	}
	ICCOMPRESSFRAMES iccf;
	ZeroMemory(&iccf, sizeof(ICCOMPRESSFRAMES));
	iccf.dwRate = SRCStreamInfo.dwRate;
	iccf.dwScale = SRCStreamInfo.dwScale;
	iccf.lQuality = CompVars.lQ;
	iccf.lDataRate = CompVars.lDataRate<<10;
	iccf.lKeyRate = CompVars.lKey;
	ICSendMessage(CompVars.hic, ICM_COMPRESS_FRAMES_INFO, (WPARAM)&iccf,
		(DWORD)sizeof(ICCOMPRESSFRAMES));
}

// ----------------------------------------------------------------------------

BOOL WriteCodecParam(FILE* fileState, COMPVARS* cvar)
{
    LONG lStateSize = 0;
    void* memState = NULL;

    lStateSize = ICGetStateSize(cvar->hic);
    memState = malloc(lStateSize);
    if (ICGetState(cvar->hic, memState, lStateSize) < 0)
    {
        ShowLastWindowsError("ICGetState failed");
        return FALSE;
    }

    fwrite(&cvar->fccHandler, sizeof(cvar->fccHandler), 1, fileState);
    fwrite(&cvar->lKey,       sizeof(cvar->lKey),       1, fileState);
    fwrite(&cvar->lQ,         sizeof(cvar->lQ),         1, fileState);
    fwrite(&cvar->lDataRate,  sizeof(cvar->lDataRate),  1, fileState);
    fwrite(&lStateSize,       sizeof(lStateSize),       1, fileState);
    fwrite(memState,          lStateSize,               1, fileState);
    free(memState);

    return TRUE;
}

// ----------------------------------------------------------------------------

BOOL ReadCodecParam(FILE* fileState, COMPVARS* cvar)
{
    LONG lStateSize = 0;
    void* memState = NULL;

    ZeroMemory(cvar,sizeof(COMPVARS));
    cvar->cbSize = sizeof(COMPVARS);
    cvar->fccType = ICTYPE_VIDEO;

    fread(&cvar->fccHandler, sizeof(cvar->fccHandler), 1, fileState);
    fread(&cvar->lKey,       sizeof(cvar->lKey),       1, fileState);
    fread(&cvar->lQ,         sizeof(cvar->lQ),         1, fileState);
    fread(&cvar->lDataRate,  sizeof(cvar->lDataRate),  1, fileState);
    fread(&lStateSize,       sizeof(lStateSize),       1, fileState);
    memState = malloc(lStateSize);
    fread(memState,          lStateSize,               1, fileState);
	cvar->dwFlags = ICMF_COMPVARS_VALID ;
    cvar->hic = ICOpen(cvar->fccType, cvar->fccHandler, ICMODE_COMPRESS);

    ICSetState(cvar->hic, memState, lStateSize);

    free(memState);

    return TRUE;
}

// ----------------------------------------------------------------------------

void SaveCodecParams(const char* filename)
{
    FILE* fileState = fopen(filename,"wb");

    // Check for 0 on the handle
    if(!fileState)
    {
        fprintf(stderr, "\nThere was a problem opening '%s'.\nTerminating...",filename);
        CleanupAndExit(1);
    }

    // Write number of passes
    fwrite(&NbPass, sizeof(NbPass), 1, fileState);
    fwrite(&current_pass, sizeof(current_pass), 1, fileState);
	
    // Write 1st pass codec state
    if(!WriteCodecParam(fileState, &CompVars))
    {
        fclose(fileState);
        CleanupAndExit(1);
    }

    // Write 2nd pass codec state if needed
    if(NbPass != 1 && !WriteCodecParam(fileState, &SecondCompVars))
    {
        fclose(fileState);
        CleanupAndExit(1);
    }

    fclose(fileState);
}

// ----------------------------------------------------------------------------

void LoadCodecParams(const char* filename)
{
    int NbPassFile = 0;
    FILE* fileState = fopen(filename,"rb");

    // Check for 0 on the handle
    if(!fileState)
    {
        fprintf(stderr, "\nThere was a problem opening '%s'.\nTerminating...",filename);
        CleanupAndExit(1);
    }

    // Read number of pass
    fread(&NbPassFile, sizeof(NbPassFile), 1, fileState);
    fread(&current_pass, sizeof(current_pass), 1, fileState);

    if(NbPassFile != NbPass)
    {
        fprintf(stderr, "\nThe file you are trying to open is a %d pass file.\nTerminating...", NbPassFile);
        CleanupAndExit(1);
    }

    // Read 1st pass codec state
    ReadCodecParam(fileState, &CompVars);
    printf("Compressor:\n");
    ShowCompressorInfo(&CompVars);

    if(NbPass != 1)
    {
        ReadCodecParam(fileState, &SecondCompVars);
		if (current_pass == 2) {
					ICCompressorFree(&SecondCompVars); // were on the second pass so we dont need both sets of compressors 
						// (we free the second even when on the second pass as when we save in the second pass
						// the second compressor is copied to the first in the main loop (trust me :)))
		}
    }
    fclose(fileState);
}

// ----------------------------------------------------------------------------

void ReadFrame()
{
    for(int foo=0;foo<VIDEO_BUFFERS;foo++)
    {
        SRCData[foo] = GlobalAllocPtr(GMEM_MOVEABLE, bmihVideoSource.biSizeImage);
    }

    for (DWORD frameNumber = StartFrame;
        frameNumber < SRCStreamInfo.dwLength && !StopRead;
        frameNumber++)
    {
        current_frame = frameNumber;

        //Make sure writing is done
        WaitForSingleObject(hSemWaitWrite, INFINITE);

        // Read source frame
        if(AVIStreamRead(pSRCVideoStream, frameNumber, 1, SRCData[frameNumber%VIDEO_BUFFERS],
            bmihVideoSource.biSizeImage, NULL, NULL) != AVIERR_OK)
        {
            ShowLastWindowsError("AVIStreamRead failed in Read thread.");
            CleanupAndExit(1);
        }
        FrameRead++;

        // Signal reading is done
		while(!ReleaseSemaphore(hSemWaitRead, 1, NULL)) {
				ShowLastWindowsError("ReleaseSemaphore(hSemWaitRead... Failed! ");
		}
    }

}

// ----------------------------------------------------------------------------

void EncodeFrame()
{
	
    void *DSTData;
    BOOL KF_flag;
    LONG DSTDataSize;
    DWORD BFrameNum=0;

    FrameRead = 0;
    FrameWritten = 0;




    if (!ICSeqCompressFrameStart(&CompVars, (LPBITMAPINFO)&bmihVideoSource))
    {
        ShowLastWindowsError("ICSeqCompressFrameStart failed");
        CleanupAndExit(1);
    }

    for (DWORD frameNumber = StartFrame;
        frameNumber < SRCStreamInfo.dwLength+BFrameNum && !StopRead;
        frameNumber++)
    {
        // Make sure reading is done
        if (frameNumber < SRCStreamInfo.dwLength)
        {
            WaitForSingleObject(hSemWaitRead, INFINITE);
        }           

        if ((DSTData = ICSeqCompressFrame(&CompVars, 0, SRCData[frameNumber%VIDEO_BUFFERS], &KF_flag, &DSTDataSize)) == NULL)
        {
            ShowLastWindowsError("ICSeqCompressFrame failed");
//			FlushBuffer();
            CleanupAndExit(1);
        }

        KFCount += KF_flag ? 1 : 0;

        // detect delay frames due to b-frames (divx 5)
        if ((DSTDataSize) == 1 && (*(byte*)DSTData==0x7f))
        {
            BFrameNum++;
        }
        else
        {
            out->writeData(DSTData,DSTDataSize, KF_flag);
        }


        // Signify writing done
		if (frameNumber < SRCStreamInfo.dwLength)
        {
			while (!ReleaseSemaphore(hSemWaitWrite, 1, NULL)) {

				ShowLastWindowsError("ReleaseSemaphore(hSemWaitWrite... Failed! ");
			}
		}
    }

    if(frameNumber == SRCStreamInfo.dwLength+BFrameNum)
    {
//        FlushBuffer();
	}

    current_frame = -1;
    BFrameNum = 0;

    // End of frame sequence compression
    ICSeqCompressFrameEnd(&CompVars);
}

// ----------------------------------------------------------------------------

void DisplayStats()
{
	unsigned int lastframe = 0;
	
    unsigned int elapsedTime = 0, remainingTime = 0;
    float fps = 0, DSTAVISizeDisplay = 0;
    static char timeBuff[16];

    __int64 freq, startTime, endTime, lasttime;
    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
    freq /= 1000;   // To have ms
    QueryPerformanceCounter((LARGE_INTEGER*)&startTime);
	lasttime = startTime;
    if(Quiet)
    {
        printf("Compressing pass %d of %d...",current_pass, NbPass);
    }
    while(!Stop && current_frame >= 0)
    {
        // Display info
        if(!Quiet)
        {
            QueryPerformanceCounter((LARGE_INTEGER*)&endTime);
	        elapsedTime = (unsigned int)((endTime - lasttime) / freq);
			if(elapsedTime > 1)
            {
                fps = (float)((FrameRead-lastframe) * 1000) / elapsedTime;
				lasttime = endTime;
				lastframe = FrameRead;
                remainingTime = (unsigned int)(((SRCStreamInfo.dwLength - FrameRead) / fps) * 1000);
                TimeMs2String(remainingTime, timeBuff);
                DSTAVISizeDisplay = (float)DSTAVISize/(1024*1024);
            }

            printf("\rPass %d/%d frame %d/%d, %.2ffps, ETA: %s, size: %.2fMB",
                current_pass, NbPass, current_frame, SRCStreamInfo.dwLength, fps,
                timeBuff, DSTAVISizeDisplay);
        }
        Sleep(300);
    }

    QueryPerformanceCounter((LARGE_INTEGER*)&endTime);

    EraseCurrentLine();
    if(NbPass == current_pass)
    {
        printf("  * Frames: %d\n",FrameWritten + StartFrame);
        printf("  * Key frames: %d\n",KFCount);
        printf("  * Size: %.3fMB\n",(float)DSTAVISize/(1024*1024));
    }
    elapsedTime = (unsigned int)((endTime - startTime) / freq);
    fps = (float)(FrameWritten * 1000.0) / elapsedTime;
    printf("Pass %d/%d, compression time %s (%.2ffps).\n", current_pass, NbPass,
        TimeMs2String(elapsedTime,timeBuff), fps);
}

// ----------------------------------------------------------------------------

void Compress()
{
    HANDLE TName[3];
    DWORD ThreadID;

	hSemWaitRead = CreateSemaphore(NULL,0,VIDEO_BUFFERS,NULL);
    hSemWaitWrite = CreateSemaphore(NULL,VIDEO_BUFFERS,VIDEO_BUFFERS,NULL);

    TName[0] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadFrame,
        NULL, 0, &ThreadID);

    TName[1]= CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)EncodeFrame,
        NULL, 0, &ThreadID);

    TName[2] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)DisplayStats,
        NULL, 0, &ThreadID);

    WaitForMultipleObjects(3, TName, TRUE, INFINITE);
    for(int foo=0;foo<VIDEO_BUFFERS;foo++)
    {
        GlobalFreePtr(SRCData[foo]);
    }       

    CloseHandle(TName[0]);
    CloseHandle(TName[1]);
    CloseHandle(TName[2]);
	CloseHandle(hSemWaitRead);
	CloseHandle(hSemWaitWrite);
}

// ============================================================================
// EOF : avs2avi.cpp
// ============================================================================
